from decimal import Decimal
from ipaddress import IPv4Address
from django.http import HttpRequest

from django_user_agents.utils import get_user_agent
from user_agents.parsers import UserAgent, parse_version
from ua_parser.user_agent_parser import ParseUserAgent

from ..const import DEFAULT_DPR
from ..dtos import InterlocutorDTO, AppDTO, OSDTO, DeviceDTO
from ..utils import stringify_version


HEADER_PREFIX = 'HTTP_'

UA_CLIENT_HINT_HEADER_PREFIX = f'{HEADER_PREFIX}SEC_CH_UA'
UA_CLIENT_HINT_GENERIC_KEY = 'generic'

DEVICE_CLIENT_HIT_HEADERS = {'DEVICE_MEMORY', 'DPR', 'VIEWPORT_WIDTH', 'WIDTH'}
NETWORK_CLIENT_HIT_HEADERS = {'DOWNLINK', 'ECT', 'RTT', 'SAVE_DATA'}

BITNESS_MAPPER = {
    '64': ('x64', 'x86_64'),
    '32': ('x32', 'x86_32'),
}
ARCH_MAPPER = {
    'x86': ('x86_32', 'x86_64'),
}

EMPTY = {None, ''}
EMPTY_FAMILIES = {'other'} | EMPTY

# Since wer'e going to use this to calculate the amount of RAM it should
# be defined the same as RAM manufactures does it. Not the truthful power of 2.
RAM_GIGABYTE = 10 ** 9


def get(source: dict, key: str, convertor: callable = None, default=None, empty=EMPTY):
    value = source.get(key, None)

    if value is None or value in empty or (isinstance(value, str) and value.lower() in empty):
        return default

    return value if convertor is None else convertor(value)


def mapper_resolver(value, mapper: dict):
    for key, values in mapper.items():
        for sub in values:
            if sub in value:
                return key


def collect_client_hints(request: HttpRequest) -> dict:
    strip = len(UA_CLIENT_HINT_HEADER_PREFIX)
    meta = request.META

    # Collecting all user agent's provided data
    client_hints = {
        (key[strip:].lstrip('_') or UA_CLIENT_HINT_GENERIC_KEY).lower(): value
        for key, value in meta.items()
        if key.startswith(UA_CLIENT_HINT_HEADER_PREFIX)
    }

    for key in DEVICE_CLIENT_HIT_HEADERS | NETWORK_CLIENT_HIT_HEADERS:
        header = HEADER_PREFIX + key
        value = meta.get(header)

        if value:
            client_hints.setdefault(key.lower(), value)

    return client_hints


def collect_app_data(ua: UserAgent, hints: dict) -> AppDTO:
    browser = ua.browser
    version = browser.version
    family = browser.family

    agent = (
        get(hints, 'full_version_list')
        or
        get(hints, UA_CLIENT_HINT_GENERIC_KEY)
        or
        get(hints, 'full_version')
    )
    if agent:
        # FIXME: This parser wouldn't work here. Should write your own.
        data = ParseUserAgent(agent)

        if data['family'].lower() not in EMPTY_FAMILIES:
            family = data['family']
            version = parse_version(
                major=data['major'], minor=data['minor'], patch=data['patch']
            )

    return AppDTO(
        family=family, version=version,
        version_string=stringify_version(version),
    )


def collect_os_data(ua: UserAgent, hints: dict) -> OSDTO:
    os = ua.os
    version = os.version
    family = os.family

    family = get(hints, 'platform', default=family, empty=EMPTY_FAMILIES)

    if 'platform_version' in hints:
        version = parse_version(
            *get(hints, 'platform_version', default='').split('.')
        )

    return OSDTO(
        family=family, version=version,
        version_string=stringify_version(version),
        arch=get(hints, 'arch') or mapper_resolver(ua.ua_string, ARCH_MAPPER),
    )


def collect_device_data(ua: UserAgent, hints: dict) -> DeviceDTO:
    device = ua.device
    model = device.model
    brand = device.brand
    family = device.family

    family = get(hints, 'platform', default=family, empty=EMPTY_FAMILIES)
    brand = brand or family
    model = get(hints, 'model', default=model)

    return DeviceDTO(
        id=(
            get(hints, 'id')
            or
            get(hints, 'identificator')
        ),
        family=family, model=model, brand=brand,
        bitness=get(hints, 'bitness') or mapper_resolver(ua.ua_string, BITNESS_MAPPER),
        memory=get(hints, 'memory', convertor=lambda x: Decimal(x) * RAM_GIGABYTE),
        dpr=get(hints, 'dpr', convertor=Decimal, default=DEFAULT_DPR),
        viewport_width=(
            get(hints, 'viewport_width', convertor=int)
            or
            get(hints, 'width', convertor=int)
        ),
    )


def resolve(request: HttpRequest) -> InterlocutorDTO:
    client_hints = collect_client_hints(request)
    ua = get_user_agent(request)
    ip = request.META.get('HTTP_X_FORWARDED_FOR')

    if ip:
        ip = ip.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')

    return InterlocutorDTO(
        ip=IPv4Address(ip) if ip else None,
        os=collect_os_data(ua, client_hints),
        app=collect_app_data(ua, client_hints),
        device=collect_device_data(ua, client_hints),

        user_agent=ua,
        client_hints=client_hints,
    )



from django.contrib import admin

from .models import InterlocutorNetwork, Interlocutor, App, OS, Device


@admin.register(App)
class AppAdmin(admin.ModelAdmin):
    list_display = 'id', 'family', 'version',
    list_filter = 'family',
    search_fields = 'id', 'family',
    date_hierarchy = 'created_at'


@admin.register(OS)
class OSAdmin(admin.ModelAdmin):
    list_display = 'id', 'family', 'arch', 'version',
    list_filter = 'family', 'arch',
    search_fields = 'id', 'family', 'arch',
    date_hierarchy = 'created_at'


@admin.register(Device)
class DeviceAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'brand', 'model', 'family', 'bitness', 'memory', 'dpr', 'viewport_width',
    )
    list_filter = 'family', 'model', 'brand', 'bitness',
    search_fields = 'id', 'family', 'model', 'brand', 'bitness',
    date_hierarchy = 'created_at'


class InterlocutorNetworkInline(admin.TabularInline):
    extra = 0
    model = InterlocutorNetwork


@admin.register(Interlocutor)
class InterlocutorAdmin(admin.ModelAdmin):
    list_display = 'id', 'os', 'device', 'app', 'outer_id',
    list_filter = 'os', 'device', 'app',
    autocomplete_fields = 'os', 'device', 'app'
    date_hierarchy = 'created_at'
    search_fields = (
        'id', 'user_agent', 'client_hints', 'outer_id',
        'device__id', 'os__id', 'app__id', 'network_connections__ip'
    )
    inlines = InterlocutorNetworkInline,

    def get_queryset(self, request):
        return super().get_queryset(request).with_relateds().distinct()



from typing import Sequence
from uuid import uuid4
from django.db import models
from django.contrib.postgres.fields import ArrayField
from django.utils.translation import pgettext_lazy

try:
    from django.db.models import JSONField
except ImportError:
    from django.contrib.postgres.fields import JSONField

from .utils import stringify_version


__all__ = 'App', 'OS', 'Device', 'Interlocutor',


def connect(items: Sequence, check=lambda x: x if x else None) -> str:
    return ' '.join(str(x) for x in items if check(x))


def VersionField(verbose_name=pgettext_lazy('wcd_device_recognizer', 'Version')):
    return ArrayField(
        models.IntegerField(), verbose_name=verbose_name,
        blank=True, default=list,
    )


class UUIDable(models.Model):
    class Meta:
        abstract = True

    id = models.UUIDField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'ID'),
        primary_key=True, default=uuid4,
    )

    created_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Created at'),
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Updated at'),
        auto_now=True
    )


class Versionable(UUIDable):
    class Meta:
        abstract = True

    version = VersionField()

    @property
    def version_string(self):
        return stringify_version(self.version)


class Familia(models.Model):
    class Meta:
        abstract = True

    family = models.CharField(
        pgettext_lazy('wcd_device_recognizer', 'Family'),
        max_length=1024, db_index=True, blank=True,
    )


class App(Versionable, Familia):
    class Meta:
        verbose_name = pgettext_lazy('wcd_device_recognizer', 'App')
        verbose_name_plural = pgettext_lazy('wcd_device_recognizer', 'Apps')

    def __str__(self):
        return connect((self.family, self.version_string))


class OS(Versionable, Familia):
    class Meta:
        verbose_name = pgettext_lazy('wcd_device_recognizer', 'OS')
        verbose_name_plural = pgettext_lazy('wcd_device_recognizer', 'OSs')

    arch = models.CharField(
        pgettext_lazy('wcd_device_recognizer', 'Architecture'),
        max_length=24, db_index=True, blank=True,
    )

    def __str__(self):
        return connect((self.family, self.version_string, self.arch))


class Device(UUIDable, Familia):
    class Meta:
        verbose_name = pgettext_lazy('wcd_device_recognizer', 'Device')
        verbose_name_plural = pgettext_lazy('wcd_device_recognizer', 'Devices')

    brand = models.CharField(
        pgettext_lazy('wcd_device_recognizer', 'Brand'),
        max_length=1024, db_index=True, blank=True,
    )
    model = models.CharField(
        pgettext_lazy('wcd_device_recognizer', 'Model'),
        max_length=1024, db_index=True, blank=True,
    )
    bitness = models.CharField(
        pgettext_lazy('wcd_device_recognizer', 'Bitness'),
        max_length=24, db_index=True, blank=True,
    )
    memory = models.DecimalField(
        pgettext_lazy('wcd_device_recognizer', 'Memory'),
        decimal_places=4, max_digits=28, null=True, blank=True,
    )
    dpr = models.DecimalField(
        pgettext_lazy('wcd_device_recognizer', 'DPR'),
        decimal_places=6, max_digits=28, default=1,
    )
    viewport_width = models.IntegerField(
        pgettext_lazy('wcd_device_recognizer', 'Viewport width'),
        null=True, blank=True,
    )

    def __str__(self):
        return connect((self.brand, self.model)) or self.family


class InterlocutorQuerySet(models.QuerySet):
    def with_relateds(self):
        return self.select_related('os', 'app', 'device')


class Interlocutor(UUIDable):
    # type: models.Manager[InterlocutorQuerySet]
    objects = InterlocutorQuerySet.as_manager()

    class Meta:
        verbose_name = pgettext_lazy('wcd_device_recognizer', 'Interlocutor')
        verbose_name_plural = pgettext_lazy('wcd_device_recognizer', 'Interlocutors')

    os = models.ForeignKey(
        OS, on_delete=models.SET_NULL, related_name='interlocutors',
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'OS'),
        null=True, blank=True,
    )
    app = models.ForeignKey(
        App, on_delete=models.SET_NULL, related_name='interlocutors',
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'App'),
        null=True, blank=True,
    )
    device = models.ForeignKey(
        Device, on_delete=models.SET_NULL, related_name='interlocutors',
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Device'),
        null=True, blank=True,
    )

    outer_id = models.TextField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Device/App identifier'),
        blank=True,
    )
    user_agent = models.TextField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'User Agent string'),
        blank=True,
    )
    client_hints = JSONField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Client hints'),
        default=dict,
    )

    def __str__(self):
        return f'#{self.id}'


class InterlocutorNetwork(UUIDable):
    class Meta:
        verbose_name = pgettext_lazy('wcd_device_recognizer', 'Interlocutor network info')
        verbose_name_plural = pgettext_lazy('wcd_device_recognizer', 'Interlocutor network infos')

    interlocutor = models.ForeignKey(
        Interlocutor, on_delete=models.CASCADE, related_name='network_connections',
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'Interlocutor'),
    )
    ip = models.GenericIPAddressField(
        verbose_name=pgettext_lazy('wcd_device_recognizer', 'IP Address'),
        null=False, blank=False,
    )

    def __str__(self):
        return f'Interlocutor #{self.interlocutor_id} connection from {self.ip}'



from dataclasses import dataclass, field
from typing import Dict, TYPE_CHECKING
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


@s('WCD_DEVICE_RECOGNIZER')
@dataclass
class Settings:
    pass


settings = Settings()



from dataclasses import dataclass, field
from ipaddress import IPv4Address
from decimal import Decimal
from typing import Dict, Optional, Tuple
from user_agents.parsers import UserAgent

from .const import DEFAULT_DPR


__all__ = 'AppDTO', 'OSDTO', 'DeviceDTO', 'InterlocutorDTO',


DEFAULT_DPR = Decimal(1)


@dataclass(frozen=True, eq=True)
class AppDTO:
    family: str
    version: Tuple[int, ...]
    version_string: str


@dataclass(frozen=True, eq=True)
class OSDTO:
    family: str
    version: Tuple[int, ...]
    version_string: str
    arch: Optional[str] = None


@dataclass(frozen=True, eq=True)
class DeviceDTO:
    id: Optional[str] = None
    family: str = 'Other'
    brand: Optional[str] = None
    model: Optional[str] = None
    bitness: Optional[str] = None
    memory: Optional[Decimal] = None
    dpr: Decimal = DEFAULT_DPR
    viewport_width: Optional[int] = None


@dataclass(frozen=True, eq=True)
class InterlocutorDTO:
    os: OSDTO
    app: AppDTO
    device: DeviceDTO

    user_agent: UserAgent
    client_hints: Dict[str, str] = field(default_factory=dict)
    ip: Optional[IPv4Address] = None



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy

from .discovery import autodiscover


__all__ = ('DeviceRecognizer',)


class DeviceRecognizer(AppConfig):
    name = 'wcd_device_recognizer'
    verbose_name = pgettext_lazy('wcd_device_recognizer', 'Device recognizer')

    def ready(self):
        autodiscover()



from typing import Iterable, List, Sequence, Tuple, Type, TypeVar
from operator import attrgetter

from django.db import models


M = TypeVar('M', bound=models.Model)


def stringify_version(version: tuple) -> str:
    return '.'.join([str(v) for v in version])


def to_list(items: Iterable) -> list:
    if isinstance(items, list):
        return items

    return list(items)


def model_bulk_get_or_create(
    model: Type[M], items: Sequence[Tuple[dict, dict]],
    internal_id_name: str = '_internal_lookup_id',
) -> List[M]:
    if len(items) == 0:
        return []

    id_map = {i + 1: value for i, value in enumerate(items)}
    q = models.Q()
    whens = []
    getid = attrgetter(internal_id_name)

    for id, (filters, _) in id_map.items():
        condition = models.Q(**filters)
        q |= condition
        whens.append(models.When(condition, then=models.Value(id)))

    existing = list(
        model.objects
        .filter(q)
        .annotate(**{internal_id_name: models.Case(
            *whens, default=None, output_field=models.IntegerField(),
        )})
    )
    diff = list(id_map.keys() - {getid(x) for x in existing})

    if len(diff) == 0:
        return existing

    blanks = (
        model(**{**id_map[id][0], **id_map[id][1]})
        for id in diff
    )
    created = model.objects.bulk_create(blanks)

    for i, x in enumerate(created):
        setattr(x, internal_id_name, diff[i])

    return sorted(existing + created, key=getid)



# Generated by Django 2.2 on 2022-07-12 16:07

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='App',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('version', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), blank=True, default=list, size=None, verbose_name='Version')),
                ('family', models.CharField(blank=True, db_index=True, max_length=1024, verbose_name='Family')),
            ],
            options={
                'verbose_name': 'App',
                'verbose_name_plural': 'Apps',
            },
        ),
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('family', models.CharField(blank=True, db_index=True, max_length=1024, verbose_name='Family')),
                ('brand', models.CharField(blank=True, db_index=True, max_length=1024, verbose_name='Brand')),
                ('model', models.CharField(blank=True, db_index=True, max_length=1024, verbose_name='Model')),
                ('bitness', models.CharField(blank=True, db_index=True, max_length=24, verbose_name='Bitness')),
                ('memory', models.DecimalField(blank=True, decimal_places=4, max_digits=28, null=True, verbose_name='Memory')),
                ('dpr', models.DecimalField(decimal_places=6, default=1, max_digits=28, verbose_name='DPR')),
                ('viewport_width', models.IntegerField(blank=True, null=True, verbose_name='Viewport width')),
            ],
            options={
                'verbose_name': 'Device',
                'verbose_name_plural': 'Devices',
            },
        ),
        migrations.CreateModel(
            name='Interlocutor',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('outer_id', models.TextField(blank=True, verbose_name='Device/App identifier')),
                ('user_agent', models.TextField(blank=True, verbose_name='User Agent string')),
                ('client_hints', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='Client hints')),
                ('app', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='interlocutors', to='wcd_device_recognizer.App', verbose_name='App')),
                ('device', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='interlocutors', to='wcd_device_recognizer.Device', verbose_name='Device')),
            ],
            options={
                'verbose_name': 'Interlocutor',
                'verbose_name_plural': 'Interlocutors',
            },
        ),
        migrations.CreateModel(
            name='OS',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('version', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), blank=True, default=list, size=None, verbose_name='Version')),
                ('family', models.CharField(blank=True, db_index=True, max_length=1024, verbose_name='Family')),
                ('arch', models.CharField(blank=True, db_index=True, max_length=24, verbose_name='Architecture')),
            ],
            options={
                'verbose_name': 'OS',
                'verbose_name_plural': 'OSs',
            },
        ),
        migrations.CreateModel(
            name='InterlocutorNetwork',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('ip', models.GenericIPAddressField(verbose_name='IP Address')),
                ('interlocutor', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='network_connections', to='wcd_device_recognizer.Interlocutor', verbose_name='Interlocutor')),
            ],
            options={
                'verbose_name': 'Interlocutor network info',
                'verbose_name_plural': 'Interlocutor network infos',
            },
        ),
        migrations.AddField(
            model_name='interlocutor',
            name='os',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='interlocutors', to='wcd_device_recognizer.OS', verbose_name='OS'),
        ),
    ]



from typing import List, Optional, Sequence, Tuple

from ..models import Interlocutor, App, OS, Device, InterlocutorNetwork
from ..dtos import InterlocutorDTO, AppDTO, OSDTO, DeviceDTO
from ..utils import model_bulk_get_or_create, to_list


def register_apps(
    items: Sequence[AppDTO]
) -> List[App]:
    return model_bulk_get_or_create(App, [
        ({'family': x.family, 'version': x.version or []}, {})
        for x in items
    ])


def register_devices(
    items: Sequence[DeviceDTO]
) -> List[Device]:
    return model_bulk_get_or_create(Device, [
        ({
            'family': x.family,
            'brand': x.brand or '',
            'model': x.model or '',
            'bitness': x.bitness or '',
            'memory': x.memory,
            'dpr': x.dpr,
            'viewport_width': x.viewport_width,
        }, {})
        for x in items
    ])


def register_oss(
    items: Sequence[OSDTO]
) -> List[OS]:
    return model_bulk_get_or_create(OS, [
        ({'family': x.family, 'version': x.version or [], 'arch': x.arch or ''}, {})
        for x in items
    ])


def register_interlocutors(
    items: Sequence[InterlocutorDTO]
) -> List[Tuple[Interlocutor, Optional[InterlocutorNetwork]]]:
    items = to_list(items)
    apps_list = list({x.app for x in items})
    oss_list = list({x.os for x in items})
    devices_list = list({x.device for x in items})

    apps = dict(zip(apps_list, register_apps(apps_list)))
    oss = dict(zip(oss_list, register_oss(oss_list)))
    devices = dict(zip(devices_list, register_devices(devices_list)))

    interlocutors = model_bulk_get_or_create(Interlocutor, [
        (
            {
                'os_id': oss[x.os].id if x.os in oss else None,
                'app_id': apps[x.app].id if x.app in apps else None,
                'device_id': devices[x.device].id if x.device in devices else None,
                'outer_id': x.device.id or '',
                'user_agent': x.user_agent.ua_string or '',
            },
            {'client_hints': x.client_hints}
        )
        for x in items
    ])
    ip_definitions = [
        (items[i].ip, x.id)
        for i, x in enumerate(interlocutors) if items[i].ip
    ]
    networks_map = {}

    if len(ip_definitions) != 0:
        networks = model_bulk_get_or_create(InterlocutorNetwork, [
            ({'ip': str(ip), 'interlocutor_id': id}, {})
            for ip, id in ip_definitions
        ])
        networks_map = {x.interlocutor_id: x for x in networks}

    return [(i, networks_map.get(i.id, None)) for i in interlocutors]
